"""Classes and functions for testing the behavior of TMTapes."""

import unittest

from automata.tm.tape import TMTape


class TestTMTape(unittest.TestCase):
    """A test class for testing all Turing machines."""

    def test_tape_copy(self) -> None:
        """Should copy TMTape."""
        tape = TMTape("0011", blank_symbol="#", current_position=0)
        new_tape = tape.copy()
        self.assertIsNot(new_tape, tape)



"""Classes and methods for working with PDA stacks."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Iterator, Sequence, Tuple


@dataclass(frozen=True)
class PDAStack:
    """A PDA stack."""

    __slots__ = ("stack",)

    stack: Tuple[str, ...]

    def __init__(self, stack: Sequence[str]) -> None:
        object.__setattr__(self, "stack", tuple(stack))

    def top(self) -> str:
        """
        Return the symbol at the top of the stack.

        Returns
        ----------
        str
            The symbol at the top of the stack. Returns the empty
            string if the stack is empty.
        """
        if self.stack:
            return self.stack[-1]

        return ""

    def pop(self) -> PDAStack:
        """
        Pop the top element from the stack and return
        the new stack.

        Returns
        ----------
        PDAStack
            A copy of the old PDAStack with the top element removed.
        """
        return self.__class__(self.stack[:-1])

    def replace(self, symbols: Sequence[str]) -> PDAStack:
        """
        Replace the top of the stack with the given symbols.

        Parameters
        ----------
        symbols : Sequence[str]
            A sequence of symbols to add to the top of the PDAStack.
            The first symbol in the given sequence becomes the new stack top.

        Returns
        ----------
        PDAStack
            A copy of the old PDAStack with the top element replaced.
        """
        return self.__class__(self.stack[:-1] + tuple(reversed(symbols)))

    def __len__(self) -> int:
        """Return the number of symbols on the stack."""
        return len(self.stack)

    def __iter__(self) -> Iterator[str]:
        """Return an iterator for the stack."""
        return iter(self.stack)

    def __getitem__(self, key: int | slice) -> str | Sequence[str]:
        """Return the stack element at the given index"""
        return self.stack[key]

    def __reversed__(self) -> Iterator[str]:
        """Return an iterator for the stack in reversed order"""
        return reversed(self.stack)

    def __repr__(self) -> str:
        """Return a string representation of the stack."""
        return "{}({})".format(self.__class__.__name__, self.stack)

    def __eq__(self, other: Any) -> bool:
        """Return True if two PDAConfiguration are equivalent"""
        if not isinstance(other, PDAStack):
            return NotImplemented

        return self.stack == other.stack



"""Classes and methods for working with all finite automata."""

from __future__ import annotations

import abc
import os
from collections import defaultdict
from typing import Any, Dict, Generator, List, Optional, Set, Tuple, Union

from automata.base.automaton import Automaton, AutomatonStateT
from automata.base.utils import (
    LayoutMethod,
    create_graph,
    create_unique_random_id,
    save_graph,
)

# Optional imports for use with visual functionality
try:
    import coloraide
    import pygraphviz as pgv
except ImportError:
    _visual_imports = False
else:
    _visual_imports = True


FAStateT = AutomatonStateT


class FA(Automaton, metaclass=abc.ABCMeta):
    """
    The `FA` class is an abstract base class from which all finite automata inherit.
    Every subclass of FA can be rendered natively inside of a Jupyter notebook
    (automatically calling `show_diagram` without any arguments) if installed with
    the `visual` optional dependency.
    """

    __slots__ = tuple()

    @staticmethod
    def _get_edge_name(symbol: str) -> str:
        return "ε" if symbol == "" else str(symbol)

    @abc.abstractmethod
    def iter_transitions(self) -> Generator[Tuple[FAStateT, FAStateT, str], None, None]:
        """
        Iterate over all transitions in the automaton. Each transition is a tuple
        of the form (from_state, to_state, symbol)
        """

        raise NotImplementedError(
            f"iter_transitions is not implemented for {self.__class__}"
        )

    def show_diagram(
        self,
        input_str: Optional[str] = None,
        path: Union[str, os.PathLike, None] = None,
        *,
        layout_method: LayoutMethod = "dot",
        horizontal: bool = True,
        reverse_orientation: bool = False,
        fig_size: Union[Tuple[float, float], Tuple[float], None] = None,
        font_size: float = 14.0,
        arrow_size: float = 0.85,
        state_separation: float = 0.5,
    ) -> pgv.AGraph:
        """
        Generates a diagram of the associated automaton.

        Parameters
        ----------
        input_str : Optional[str], default: None
            String consisting of input symbols. If set, will add processing of
            the input string to the diagram.
        path : Union[str, os.PathLike, None], default: None
            Path to output file. If None, the output will not be saved.
        horizontal : bool, default: True
            Direction of node layout in the output graph.
        reverse_orientation : bool, default: False
            Reverse direction of node layout in the output graph.
        fig_size : Union[Tuple[float, float], Tuple[float], None], default: None
            Figure size.
        font_size : float, default: 14.0
            Font size in the output graph.
        arrow_size : float, default: 0.85
            Arrow size in the output graph.
        state_separation : float, default: 0.5
            Distance between nodes in the output graph.

        Returns
        ------
        AGraph
            A diagram of the given automaton.
        """

        if not _visual_imports:
            raise ImportError(
                "Missing visualization packages; "
                "please install coloraide and pygraphviz."
            )

        # Defining the graph.
        graph = create_graph(
            horizontal, reverse_orientation, fig_size, state_separation
        )

        font_size_str = str(font_size)
        arrow_size_str = str(arrow_size)

        # create unique id to avoid colliding with other states
        null_node = create_unique_random_id()

        graph.add_node(
            null_node,
            label="",
            tooltip=".",
            shape="point",
            fontsize=font_size_str,
        )
        initial_node = self._get_state_name(self.initial_state)
        graph.add_edge(
            null_node,
            initial_node,
            tooltip="->" + initial_node,
            arrowsize=arrow_size_str,
        )

        nonfinal_states = map(self._get_state_name, self.states - self.final_states)
        final_states = map(self._get_state_name, self.final_states)
        graph.add_nodes_from(nonfinal_states, shape="circle", fontsize=font_size_str)
        graph.add_nodes_from(final_states, shape="doublecircle", fontsize=font_size_str)

        is_edge_drawn = defaultdict(lambda: False)
        if input_str is not None:
            input_path, is_accepted = self._get_input_path(input_str=input_str)

            start_color = coloraide.Color("#ff0")
            end_color = (
                coloraide.Color("#0f0") if is_accepted else coloraide.Color("#f00")
            )
            interpolation = coloraide.Color.interpolate(
                [start_color, end_color], space="srgb"
            )

            # find all transitions in the finite state machine with traversal.
            for transition_index, (from_state, to_state, symbol) in enumerate(
                input_path, start=1
            ):
                color = interpolation(transition_index / len(input_path))
                label = self._get_edge_name(symbol)

                is_edge_drawn[from_state, to_state, symbol] = True
                graph.add_edge(
                    self._get_state_name(from_state),
                    self._get_state_name(to_state),
                    label=f"<{label} <b>[<i>#{transition_index}</i>]</b>>",
                    arrowsize=arrow_size_str,
                    fontsize=font_size_str,
                    color=color.to_string(hex=True),
                    penwidth="2.5",
                )

        edge_labels = defaultdict(list)
        for from_state, to_state, symbol in self.iter_transitions():
            if is_edge_drawn[from_state, to_state, symbol]:
                continue

            from_node = self._get_state_name(from_state)
            to_node = self._get_state_name(to_state)
            label = self._get_edge_name(symbol)
            edge_labels[from_node, to_node].append(label)

        for (from_node, to_node), labels in edge_labels.items():
            graph.add_edge(
                from_node,
                to_node,
                label=",".join(sorted(labels)),
                arrowsize=arrow_size_str,
                fontsize=font_size_str,
            )

        # Set layout
        graph.layout(prog=layout_method)

        # Write diagram to file
        if path is not None:
            save_graph(graph, path)

        return graph

    @abc.abstractmethod
    def _get_input_path(
        self, input_str: str
    ) -> Tuple[List[Tuple[FAStateT, FAStateT, str]], bool]:
        """Calculate the path taken by input."""

        raise NotImplementedError(
            f"_get_input_path is not implemented for {self.__class__}"
        )

    def _repr_mimebundle_(
        self, *args: Any, **kwargs: Any
    ) -> Dict[str, Union[bytes, str]]:
        return self.show_diagram()._repr_mimebundle_(*args, **kwargs)

    @staticmethod
    def _add_new_state(state_set: Set[FAStateT], start: int = 0) -> int:
        """Adds new state to the state set and returns it"""
        new_state = start
        while new_state in state_set:
            new_state += 1

        state_set.add(new_state)

        return new_state



"""Classes and functions for testing the behavior of NPDAs."""

import os

from frozendict import frozendict

import automata.base.exceptions as exceptions
import automata.pda.exceptions as pda_exceptions
import tests.test_pda as test_pda
from automata.pda.configuration import PDAConfiguration
from automata.pda.npda import NPDA
from automata.pda.stack import PDAStack


class TestNPDA(test_pda.TestPDA):
    """A test class for testing nondeterministic finite automata."""

    def test_init_npda(self) -> None:
        """Should copy NPDA if passed into NPDA constructor."""
        new_npda = self.npda.copy()
        self.assertIsNot(new_npda, self.npda)

    def test_init_npda_missing_formal_params(self) -> None:
        """Should raise an error if formal NPDA parameters are missing."""
        with self.assertRaises(TypeError):
            NPDA(  # type: ignore
                states={"q0", "q1", "q2"},
                input_symbols={"a", "b"},
                initial_state="q0",
                final_states={"q0"},
            )

    def test_init_npda_no_acceptance_mode(self) -> None:
        """Should create a new NPDA."""
        new_npda = NPDA(
            states={"q0"},
            input_symbols={"a", "b"},
            stack_symbols={"#"},
            transitions={
                "q0": {
                    "a": {"#": {("q0", "")}},
                }
            },
            initial_state="q0",
            initial_stack_symbol="#",
            final_states={"q0"},
        )
        self.assertEqual(new_npda.acceptance_mode, "both")

    def test_init_npda_invalid_acceptance_mode(self) -> None:
        """Should raise an error if the NPDA has an invalid acceptance mode."""
        with self.assertRaises(pda_exceptions.InvalidAcceptanceModeError):
            NPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": {("q0", "")}}}},
                initial_state="q0",
                initial_stack_symbol="#",
                final_states={"q0"},
                acceptance_mode="foo",  # type: ignore
            )

    def test_npda_immutable_attr_set(self) -> None:
        with self.assertRaises(AttributeError):
            self.npda.states = set()

    def test_npda_immutable_attr_del(self) -> None:
        with self.assertRaises(AttributeError):
            del self.npda.states

    def test_npda_immutable_dict(self) -> None:
        """Should create an NPDA whose contents are fully immutable/hashable"""
        self.assertIsInstance(hash(frozendict(self.npda.input_parameters)), int)

    def test_validate_invalid_input_symbol(self) -> None:
        """Should raise error if a transition has an invalid input symbol."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            NPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"c": {"#": {("q0", "")}}}},
                initial_state="q0",
                initial_stack_symbol="#",
                final_states={"q0"},
            )

    def test_validate_invalid_stack_symbol(self) -> None:
        """Should raise error if a transition has an invalid stack symbol."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            NPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"@": {("q0", "")}}}},
                initial_state="q0",
                initial_stack_symbol="#",
                final_states={"q0"},
            )

    def test_validate_invalid_initial_state(self) -> None:
        """Should raise error if the initial state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            NPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": {("q0", "")}}}},
                initial_state="q0",
                initial_stack_symbol="#",
                final_states={"q1"},
            )

    def test_validate_invalid_initial_stack_symbol(self) -> None:
        """Should raise error if the initial stack symbol is invalid."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            NPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": {("q0", "")}}}},
                initial_state="q0",
                initial_stack_symbol="@",
                final_states={"q0"},
            )

    def test_validate_invalid_final_state(self) -> None:
        """Should raise error if the final state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            NPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": {("q0", "")}}}},
                initial_state="q0",
                initial_stack_symbol="#",
                final_states={"q1"},
            )

    def test_validate_invalid_final_state_non_str(self) -> None:
        """Should raise InvalidStateError even for non-string final states."""
        with self.assertRaises(exceptions.InvalidStateError):
            NPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": {("q0", "")}}}},
                initial_state="q0",
                initial_stack_symbol="#",
                final_states={1},
            )

    def test_read_input_valid_accept_by_final_state(self) -> None:
        """Should return correct config if NPDA accepts by final state."""
        self.assertEqual(
            self.npda.read_input("abaaba"),
            {PDAConfiguration("q2", "", PDAStack(["#"]))},
        )

    def test_read_input_invalid_accept_by_final_state(self) -> None:
        """Should not accept by final state if NPDA accepts by empty stack."""
        npda = NPDA(
            states={"q0", "q1", "q2"},
            input_symbols={"a", "b"},
            stack_symbols={"A", "B", "#"},
            transitions={
                "q0": {
                    "": {
                        "#": {("q2", "#")},
                    },
                    "a": {
                        "#": {("q0", ("A", "#"))},
                        "A": {
                            ("q0", ("A", "A")),
                            ("q1", ""),
                        },
                        "B": {("q0", ("A", "B"))},
                    },
                    "b": {
                        "#": {("q0", ("B", "#"))},
                        "A": {("q0", ("B", "A"))},
                        "B": {
                            ("q0", ("B", "B")),
                            ("q1", ""),
                        },
                    },
                },
                "q1": {
                    "": {"#": {("q2", "#")}},
                    "a": {"A": {("q1", "")}},
                    "b": {"B": {("q1", "")}},
                },
            },
            initial_state="q0",
            initial_stack_symbol="#",
            final_states={"q2"},
            acceptance_mode="empty_stack",
        )
        with self.assertRaises(exceptions.RejectionException):
            (npda.read_input("abaaba"),)

    def test_read_input_valid_accept_by_empty_stack(self) -> None:
        """Should return correct config if NPDA accepts by empty stack."""
        npda = NPDA(
            states={"q0", "q1", "q2"},
            input_symbols={"a", "b"},
            stack_symbols={"A", "B", "#"},
            transitions={
                "q0": {
                    "": {
                        "#": {("q2", "#")},
                    },
                    "a": {
                        "#": {("q0", ("A", "#"))},
                        "A": {
                            ("q0", ("A", "A")),
                            ("q1", ""),
                        },
                        "B": {("q0", ("A", "B"))},
                    },
                    "b": {
                        "#": {("q0", ("B", "#"))},
                        "A": {("q0", ("B", "A"))},
                        "B": {
                            ("q0", ("B", "B")),
                            ("q1", ""),
                        },
                    },
                },
                "q1": {
                    "": {"#": {("q2", "#")}},
                    "a": {"A": {("q1", "")}},
                    "b": {"B": {("q1", "")}},
                },
                "q2": {"": {"#": {("q2", "")}}},
            },
            initial_state="q0",
            initial_stack_symbol="#",
            final_states=set(),
            acceptance_mode="empty_stack",
        )
        self.assertEqual(
            npda.read_input("abaaba"), {PDAConfiguration("q2", "", PDAStack([]))}
        )

    def test_read_input_invalid_accept_by_empty_stack(self) -> None:
        """Should not accept by empty stack if NPDA accepts by final state."""
        npda = NPDA(
            states={"q0", "q1", "q2", "q3"},
            input_symbols={"a", "b"},
            stack_symbols={"A", "B", "#"},
            transitions={
                "q0": {
                    "": {
                        "#": {("q2", "#")},
                    },
                    "a": {
                        "#": {("q0", ("A", "#"))},
                        "A": {
                            ("q0", ("A", "A")),
                            ("q1", ""),
                        },
                        "B": {("q0", ("A", "B"))},
                    },
                    "b": {
                        "#": {("q0", ("B", "#"))},
                        "A": {("q0", ("B", "A"))},
                        "B": {
                            ("q0", ("B", "B")),
                            ("q1", ""),
                        },
                    },
                },
                "q1": {
                    "": {"#": {("q3", "")}},
                    "a": {"A": {("q1", "")}},
                    "b": {"B": {("q1", "")}},
                },
            },
            initial_state="q0",
            initial_stack_symbol="#",
            final_states={"q2"},
            acceptance_mode="final_state",
        )
        with self.assertRaises(exceptions.RejectionException):
            npda.read_input("abaaba")

    def test_read_input_valid_consecutive_lambda_transitions(self) -> None:
        """Should follow consecutive lambda transitions when validating."""
        npda = NPDA(
            states={"q0", "q1", "q2", "q3", "q4"},
            input_symbols={"a", "b"},
            stack_symbols={"A", "B", "#"},
            transitions={
                "q0": {
                    "": {
                        "#": {("q2", "#")},
                    },
                    "a": {
                        "#": {("q0", ("A", "#"))},
                        "A": {
                            ("q0", ("A", "A")),
                            ("q1", ""),
                        },
                        "B": {("q0", ("A", "B"))},
                    },
                    "b": {
                        "#": {("q0", ("B", "#"))},
                        "A": {("q0", ("B", "A"))},
                        "B": {
                            ("q0", ("B", "B")),
                            ("q1", ""),
                        },
                    },
                },
                "q1": {
                    "": {"#": {("q2", "#")}},
                    "a": {"A": {("q1", "")}},
                    "b": {"B": {("q1", "")}},
                },
                "q2": {"": {"#": {("q3", "#")}}},
                "q3": {"": {"#": {("q4", "#")}}},
            },
            initial_state="q0",
            initial_stack_symbol="#",
            final_states={"q4"},
            acceptance_mode="final_state",
        )
        self.assertEqual(
            npda.read_input("abaaba"), {PDAConfiguration("q4", "", PDAStack(["#"]))}
        )

    def test_read_input_rejected_accept_by_final_state(self) -> None:
        """Should reject strings if NPDA accepts by final state."""
        with self.assertRaises(exceptions.RejectionException):
            self.npda.read_input("aaba")

    def test_read_input_rejected_accept_by_empty_stack(self) -> None:
        """Should reject strings if NPDA accepts by empty stack."""
        npda = NPDA(
            states={"q0", "q1", "q2"},
            input_symbols={"a", "b"},
            stack_symbols={"A", "B", "#"},
            transitions={
                "q0": {
                    "": {
                        "#": {("q2", "#")},
                    },
                    "a": {
                        "#": {("q0", ("A", "#"))},
                        "A": {
                            ("q0", ("A", "A")),
                            ("q1", ""),
                        },
                        "B": {("q0", ("A", "B"))},
                    },
                    "b": {
                        "#": {("q0", ("B", "#"))},
                        "A": {("q0", ("B", "A"))},
                        "B": {
                            ("q0", ("B", "B")),
                            ("q1", ""),
                        },
                    },
                },
                "q1": {
                    "": {"#": {("q2", "#")}},
                    "a": {"A": {("q1", "")}},
                    "b": {"B": {("q1", "")}},
                },
                "q2": {"": {"#": {("q2", "")}}},
            },
            initial_state="q0",
            initial_stack_symbol="#",
            final_states=set(),
            acceptance_mode="final_state",
        )
        with self.assertRaises(exceptions.RejectionException):
            npda.read_input("aaba")

    def test_read_input_rejected_undefined_transition(self) -> None:
        """Should reject strings which lead to an undefined transition."""
        with self.assertRaises(exceptions.RejectionException):
            self.npda.read_input("01")

    def test_accepts_input_true(self) -> None:
        """Should return True if NPDA input is accepted."""
        self.assertTrue(self.npda.accepts_input("abaaba"))

    def test_accepts_input_false(self) -> None:
        """Should return False if NPDA input is rejected."""
        self.assertFalse(self.npda.accepts_input("aaba"))

    def test_show_diagram(self) -> None:
        """Should construct the diagram for a NPDA"""
        graph = self.npda.show_diagram()
        node_names = {node.get_name() for node in graph.nodes()}
        self.assertTrue(set(self.npda.states).issubset(node_names))
        self.assertEqual(len(self.npda.states) + 1, len(node_names))

        for state in self.npda.states:
            node = graph.get_node(state)
            expected_shape = (
                "doublecircle" if state in self.npda.final_states else "circle"
            )
            self.assertEqual(node.attr["shape"], expected_shape)

        expected_transitions = {
            ("q0", frozenset({"a, A | ε", "b, B | ε"}), "q1"),
            (
                "q0",
                frozenset(
                    {
                        "ε, # | #",
                    }
                ),
                "q2",
            ),
            (
                "q0",
                frozenset(
                    {
                        "a, # | A#",
                        "a, A | AA",
                        "a, B | AB",
                        "b, # | B#",
                        "b, A | BA",
                        "b, B | BB",
                    }
                ),
                "q0",
            ),
            ("q1", frozenset({"a, A | ε", "b, B | ε"}), "q1"),
            (
                "q1",
                frozenset(
                    {
                        "ε, # | #",
                    }
                ),
                "q2",
            ),
        }

        seen_transitions = {
            (
                edge[0],
                frozenset(map(str.strip, edge.attr["label"].split("\n"))),
                edge[1],
            )
            for edge in graph.edges()
        }

        self.assertTrue(expected_transitions.issubset(seen_transitions))
        self.assertEqual(len(expected_transitions) + 1, len(seen_transitions))

        source, symbol, dest = list(seen_transitions - expected_transitions)[0]
        self.assertEqual(symbol, frozenset({""}))
        self.assertEqual(dest, self.npda.initial_state)
        self.assertTrue(source not in self.npda.states)

    def test_show_diagram_read_input_machine_only(self) -> None:
        """
        Should construct the diagram with machine only for a NPDA reading input.
        """
        input_strings = ["abba", "aabbaa", "aabaabaa", ""]

        for input_string in input_strings:
            graph = self.npda.show_diagram(input_str=input_string, with_stack=False)

            # Get edges corresponding to input path
            colored_edges = [
                edge for edge in graph.edges() if "color" in dict(edge.attr)
            ]

            edge_pairs = [
                (edge[0].state, edge[1].state)
                for edge in self.npda._get_input_path(input_string)[0]
            ]
            self.assertEqual(edge_pairs, colored_edges)

    def test_show_diagram_read_input_stack_only(self) -> None:
        """
        Should construct the diagram with stack only for a NPDA reading input.
        """
        input_strings = ["abba", "aabbaa", "aabaabaa", ""]

        for input_string in input_strings:
            graph = self.npda.show_diagram(input_str=input_string, with_machine=False)

            # Get edges corresponding to input path
            colored_edges = [
                edge for edge in graph.edges() if "color" in dict(edge.attr)
            ]

            input_path = self.npda._get_input_path(input_string)[0]

            edge_pairs = [(str(edge[0]), str(edge[1])) for edge in input_path]

            # Get stack content corresponding to input path
            nodes = [node.attr["label"] for node in graph.nodes()]

            stack_content = [
                " | " + " | ".join(reversed(edge[0].stack)) for edge in input_path
            ] + [" | " + " | ".join(reversed(input_path[-1][1].stack))]

            self.assertEqual(nodes, stack_content)
            self.assertEqual(edge_pairs, colored_edges)

    def test_show_diagram_write_file(self) -> None:
        """
        Should construct the diagram for a NPDA
        and write it to the specified file.
        """
        diagram_path = os.path.join(self.temp_dir_path, "test_npda.png")
        try:
            os.remove(diagram_path)
        except OSError:
            pass
        self.assertFalse(os.path.exists(diagram_path))
        self.npda.show_diagram(path=diagram_path)
        self.assertTrue(os.path.exists(diagram_path))
        os.remove(diagram_path)

    def test_show_diagram_orientations(self) -> None:
        graph = self.npda.show_diagram()
        self.assertEqual(graph.graph_attr["rankdir"], "LR")
        graph = self.npda.show_diagram(horizontal=False)
        self.assertEqual(graph.graph_attr["rankdir"], "TB")
        graph = self.npda.show_diagram(reverse_orientation=True)
        self.assertEqual(graph.graph_attr["rankdir"], "RL")
        graph = self.npda.show_diagram(horizontal=False, reverse_orientation=True)
        self.assertEqual(graph.graph_attr["rankdir"], "BT")

    def test_show_diagram_fig_size(self) -> None:
        """
        Testing figure size. Just need to make sure it matches the input
        (the library handles the rendering).
        """
        graph = self.npda.show_diagram(fig_size=(1.1, 2))
        self.assertEqual(graph.graph_attr["size"], "1.1, 2")

        graph = self.npda.show_diagram(fig_size=(3.3,))
        self.assertEqual(graph.graph_attr["size"], "3.3")



"""
A set of tools for working with regular expressions. Can recognize regular expressions
over the alphabet of unicode characters and subsets of these, excluding blanks.

A regular expression with the following operations only are supported in this library:

- `*`: Kleene star operation, language repeated zero or more times. Ex: `a*`,`(ab)*`
- `+`: Kleene plus operation, language repeated one or more times. Ex: `a+`,`(ab)+`
- `?`: Language repeated zero or one time. Ex: `a?`
- Concatenation. Ex: `abcd`
- `|`: Union. Ex: `a|b`
- `&`: Intersection. Ex: `a&b`
- `.`: Wildcard. Ex: `a.b`
- `^`: Shuffle. Ex: `a^b`
- `{}`: Quantifiers expressing finite repetitions. Ex: `a{1,2}`,`a{3,}`
- `()`: The empty string.
- `(...)`: Grouping.

This is similar to the Python `re` module, but this library does not support any special
characters other than those given above. All regular languages can be written with
these.
"""

from itertools import count
from typing import AbstractSet, Optional

import automata.base.exceptions as exceptions
from automata.fa.nfa import NFA
from automata.regex.parser import RESERVED_CHARACTERS, get_regex_lexer, validate_tokens


def _validate(regex: str) -> bool:
    """Return True if the regular expression is valid"""

    try:
        validate(regex)
    except exceptions.InvalidRegexError:
        return False

    return True


def validate(regex: str) -> None:
    """
    Raises an exception if the input regular expression is invalid.

    Raises
    ------
    InvalidRegexError
        Raised if the regex given as input is not well defined.
    """
    input_symbols = set(regex) - RESERVED_CHARACTERS

    validate_tokens(get_regex_lexer(input_symbols, count(0)).lex(regex))


def isequal(
    re1: str, re2: str, *, input_symbols: Optional[AbstractSet[str]] = None
) -> bool:
    """
    Whether both regular expressions are equivalent.

    Parameters
    ----------
    re1 : str
        The first regular expression as a string.
    re2 : str
        The second regular expression as a string.
    input_symbols : Optional[AbstractSet[str]], default: None
        The set of input symbols when doing the comparison. Defaults to
        all ascii letters and digits.

    Returns
    ------
    bool
        Whether the regular expressions are equivalent.
    """

    nfa1 = NFA.from_regex(re1, input_symbols=input_symbols)
    nfa2 = NFA.from_regex(re2, input_symbols=input_symbols)

    return nfa1 == nfa2


def issubset(
    re1: str, re2: str, *, input_symbols: Optional[AbstractSet[str]] = None
) -> bool:
    """
    Whether re1 is a subset of re2.

    Parameters
    ----------
    re1 : str
        The first regular expression as a string.
    re2 : str
        The second regular expression as a string.
    input_symbols : Optional[AbstractSet[str]], default: None
        The set of input symbols when doing the comparison. Defaults to
        all ascii letters and digits.

    Returns
    ------
    bool
        True if re1 is a subset of re2.
    """

    nfa1 = NFA.from_regex(re1, input_symbols=input_symbols)
    nfa2 = NFA.from_regex(re2, input_symbols=input_symbols)

    return nfa1.union(nfa2) == nfa2


def issuperset(
    re1: str, re2: str, *, input_symbols: Optional[AbstractSet[str]] = None
) -> bool:
    """
    Whether re1 is a superset of re2.

    Parameters
    ----------
    re1 : str
        The first regular expression as a string.
    re2 : str
        The second regular expression as a string.
    input_symbols : Optional[AbstractSet[str]], default: None
        The set of input symbols when doing the comparison. Defaults to
        all ascii letters and digits.

    Returns
    ------
    bool
        True if re1 is a superset of re2.
    """

    nfa1 = NFA.from_regex(re1, input_symbols=input_symbols)
    nfa2 = NFA.from_regex(re2, input_symbols=input_symbols)

    return nfa1.union(nfa2) == nfa1



"""Classes and functions for testing the behavior of DPDAs."""

import os

from frozendict import frozendict

import automata.base.exceptions as exceptions
import automata.pda.exceptions as pda_exceptions
import tests.test_pda as test_pda
from automata.pda.configuration import PDAConfiguration
from automata.pda.dpda import DPDA
from automata.pda.stack import PDAStack


class TestDPDA(test_pda.TestPDA):
    """A test class for testing deterministic finite automata."""

    def test_init_dpda(self) -> None:
        """Should copy DPDA if passed into DPDA constructor."""
        new_dpda = self.dpda.copy()
        self.assertIsNot(new_dpda, self.dpda)

    def test_init_dpda_missing_formal_params(self) -> None:
        """Should raise an error if formal DPDA parameters are missing."""
        with self.assertRaises(TypeError):
            DPDA(  # type: ignore
                states={"q0", "q1", "q2"},
                input_symbols={"a", "b"},
                initial_state="q0",
                final_states={"q0"},
            )

    def test_init_dpda_no_acceptance_mode(self) -> None:
        """Should create a new DPDA."""
        new_dpda = DPDA(
            states={"q0"},
            input_symbols={"a", "b"},
            stack_symbols={"#"},
            transitions={"q0": {"a": {"#": ("q0", "")}}},
            initial_state="q0",
            initial_stack_symbol="#",
            final_states={"q0"},
        )
        self.assertEqual(new_dpda.acceptance_mode, "both")

    def test_init_dpda_invalid_acceptance_mode(self) -> None:
        """Should raise an error if the NPDA has an invalid acceptance mode."""
        with self.assertRaises(pda_exceptions.InvalidAcceptanceModeError):
            DPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": ("q0", "")}}},
                initial_state="q0",
                initial_stack_symbol="#",
                final_states={"q0"},
                acceptance_mode="foo",  # type: ignore
            )

    def test_dpda_immutable_attr_set(self) -> None:
        with self.assertRaises(AttributeError):
            self.dpda.states = set()

    def test_dpda_immutable_attr_del(self) -> None:
        with self.assertRaises(AttributeError):
            del self.dpda.states

    def test_dpda_immutable_dict(self) -> None:
        """Should create a DPDA whose contents are fully immutable/hashable"""
        self.assertIsInstance(hash(frozendict(self.dpda.input_parameters)), int)

    def test_validate_invalid_input_symbol(self) -> None:
        """Should raise error if a transition has an invalid input symbol."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            DPDA(
                states={"q0", "q1", "q2", "q3"},
                input_symbols={"a", "b"},
                stack_symbols={"0", "1"},
                transitions={
                    "q0": {"a": {"0": ("q1", ("1", "0"))}},
                    "q1": {
                        "a": {"1": ("q1", ("1", "1"))},
                        "b": {"1": ("q2", "")},
                        "c": {"1": ("q2", "")},
                    },
                    "q2": {"b": {"1": ("q2", "")}, "": {"0": ("q3", ("0",))}},
                },
                initial_state="q0",
                initial_stack_symbol="0",
                final_states={"q3"},
                acceptance_mode="final_state",
            )

    def test_validate_invalid_stack_symbol(self) -> None:
        """Should raise error if a transition has an invalid stack symbol."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            DPDA(
                states={"q0", "q1", "q2", "q3"},
                input_symbols={"a", "b"},
                stack_symbols={"0", "1"},
                transitions={
                    "q0": {"a": {"0": ("q1", ("1", "0"))}},
                    "q1": {
                        "a": {"1": ("q1", ("1", "1")), "2": ("q1", ("1", "1"))},
                        "b": {"1": ("q2", "")},
                    },
                    "q2": {"b": {"1": ("q2", "")}, "": {"0": ("q3", ("0",))}},
                },
                initial_state="q0",
                initial_stack_symbol="0",
                final_states={"q3"},
                acceptance_mode="final_state",
            )

    def test_validate_nondeterminism(self) -> None:
        """Should raise error if DPDA exhibits nondeterminism."""
        with self.assertRaises(pda_exceptions.NondeterminismError):
            DPDA(
                states={"q0", "q1", "q2", "q3"},
                input_symbols={"a", "b"},
                stack_symbols={"0", "1"},
                transitions={
                    "q0": {"a": {"0": ("q1", ("1", "0"))}},
                    "q1": {"a": {"1": ("q1", ("1", "1"))}, "b": {"1": ("q2", "")}},
                    "q2": {
                        "b": {"0": ("q2", "0"), "1": ("q2", "")},
                        "": {"0": ("q3", ("0",))},
                    },
                },
                initial_state="q0",
                initial_stack_symbol="0",
                final_states={"q3"},
                acceptance_mode="final_state",
            )

    def test_validate_invalid_initial_state(self) -> None:
        """Should raise error if the initial state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            DPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": ("q0", "")}}},
                initial_state="q1",
                initial_stack_symbol="#",
                final_states={"q0"},
            )

    def test_validate_invalid_initial_stack_symbol(self) -> None:
        """Should raise error if the initial stack symbol is invalid."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            DPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": ("q0", "")}}},
                initial_state="q0",
                initial_stack_symbol="2",
                final_states={"q0"},
            )

    def test_validate_invalid_final_state(self) -> None:
        """Should raise error if the final state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            DPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": ("q0", "")}}},
                initial_state="q0",
                initial_stack_symbol="#",
                final_states={"q1"},
            )

    def test_validate_invalid_final_state_non_str(self) -> None:
        """Should raise InvalidStateError even for non-string final states."""
        with self.assertRaises(exceptions.InvalidStateError):
            DPDA(
                states={"q0"},
                input_symbols={"a", "b"},
                stack_symbols={"#"},
                transitions={"q0": {"a": {"#": ("q0", "")}}},
                initial_state="q0",
                initial_stack_symbol="#",
                final_states={1},
            )

    def test_read_input_valid_accept_by_final_state(self) -> None:
        """Should return correct config if DPDA accepts by final state."""
        self.assertEqual(
            self.dpda.read_input("aabb"), PDAConfiguration("q3", "", PDAStack(["0"]))
        )

    def test_read_input_invalid_accept_by_final_state(self) -> None:
        """Should not accept by final state if DPDA accepts by empty stack."""
        dpda = DPDA(
            states={"q0", "q1", "q2", "q3"},
            input_symbols={"a", "b"},
            stack_symbols={"0", "1"},
            transitions={
                "q0": {"a": {"0": ("q1", ("1", "0"))}},
                "q1": {"a": {"1": ("q1", ("1", "1"))}, "b": {"1": ("q2", "")}},
                "q2": {"b": {"1": ("q2", "")}, "": {"0": ("q3", ("0",))}},
            },
            initial_state="q0",
            initial_stack_symbol="0",
            final_states={"q3"},
            acceptance_mode="empty_stack",
        )
        with self.assertRaises(exceptions.RejectionException):
            dpda.read_input("aabb")

    def test_read_input_valid_accept_by_empty_stack(self) -> None:
        """Should return correct config if DPDA accepts by empty stack."""
        dpda = DPDA(
            states={"q0", "q1", "q2", "q3"},
            input_symbols={"a", "b"},
            stack_symbols={"0", "1"},
            transitions={
                "q0": {"a": {"0": ("q1", ("1", "0"))}},
                "q1": {"a": {"1": ("q1", ("1", "1"))}, "b": {"1": ("q2", "")}},
                "q2": {"b": {"1": ("q2", "")}, "": {"0": ("q2", "")}},
            },
            initial_state="q0",
            initial_stack_symbol="0",
            final_states={"q3"},
            acceptance_mode="empty_stack",
        )
        self.assertEqual(
            dpda.read_input("aabb"), PDAConfiguration("q2", "", PDAStack([]))
        )

    def test_read_input_invalid_accept_by_empty_stack(self) -> None:
        """Should not accept by empty stack if DPDA accepts by final state."""
        dpda = DPDA(
            states={"q0", "q1", "q2", "q3"},
            input_symbols={"a", "b"},
            stack_symbols={"0", "1"},
            transitions={
                "q0": {"a": {"0": ("q1", ("1", "0"))}},
                "q1": {"a": {"1": ("q1", ("1", "1"))}, "b": {"1": ("q2", "")}},
                "q2": {"b": {"1": ("q2", "")}, "": {"0": ("q2", "")}},
            },
            initial_state="q0",
            initial_stack_symbol="0",
            final_states={"q3"},
            acceptance_mode="final_state",
        )
        with self.assertRaises(exceptions.RejectionException):
            dpda.read_input("aabb")

    def test_read_input_valid_consecutive_lambda_transitions(self) -> None:
        """Should follow consecutive lambda transitions when validating."""
        dpda = DPDA(
            states={"q0", "q1", "q2", "q3", "q4"},
            input_symbols={"a", "b"},
            stack_symbols={"0", "1"},
            transitions={
                "q0": {"a": {"0": ("q1", ("1", "0"))}},
                "q1": {"a": {"1": ("q1", ("1", "1"))}, "b": {"1": ("q2", "")}},
                "q2": {"b": {"1": ("q2", "")}, "": {"0": ("q3", ("0",))}},
                "q3": {"": {"0": ("q4", ("0",))}},
            },
            initial_state="q0",
            initial_stack_symbol="0",
            final_states={"q4"},
            acceptance_mode="final_state",
        )
        self.assertEqual(
            dpda.read_input("aabb"), PDAConfiguration("q4", "", PDAStack(["0"]))
        )

    def test_read_input_rejected_accept_by_final_state(self) -> None:
        """Should reject strings if DPDA accepts by final state."""
        with self.assertRaises(exceptions.RejectionException):
            self.dpda.read_input("aab")

    def test_read_input_rejected_accept_by_empty_stack(self) -> None:
        """Should reject strings if DPDA accepts by empty stack."""
        dpda = DPDA(
            states={"q0", "q1", "q2", "q3"},
            input_symbols={"a", "b"},
            stack_symbols={"0", "1"},
            transitions={
                "q0": {"a": {"0": ("q1", ("1", "0"))}},
                "q1": {"a": {"1": ("q1", ("1", "1"))}, "b": {"1": ("q2", "")}},
                "q2": {"b": {"1": ("q2", "")}, "": {"0": ("q2", "")}},
            },
            initial_state="q0",
            initial_stack_symbol="0",
            final_states={"q3"},
            acceptance_mode="final_state",
        )
        with self.assertRaises(exceptions.RejectionException):
            dpda.read_input("aab")

    def test_read_input_rejected_undefined_transition(self) -> None:
        """Should reject strings which lead to an undefined transition."""
        with self.assertRaises(exceptions.RejectionException):
            self.dpda.read_input("01")

    def test_accepts_input_true(self) -> None:
        """Should return False if DPDA input is not accepted."""
        self.assertTrue(self.dpda.accepts_input("aabb"))

    def test_accepts_input_false(self) -> None:
        """Should return False if DPDA input is rejected."""
        self.assertFalse(self.dpda.accepts_input("aab"))

    def test_empty_dpda(self) -> None:
        """Should accept an empty input if the DPDA is empty."""
        dpda = DPDA(
            states={"q0"},
            input_symbols=set(),
            stack_symbols={"0"},
            transitions=dict(),
            initial_state="q0",
            initial_stack_symbol="0",
            final_states={"q0"},
            acceptance_mode="both",
        )
        self.assertTrue(dpda.accepts_input(""))

    def test_show_diagram(self) -> None:
        """Should construct the diagram for a DPDA"""
        graph = self.dpda.show_diagram()
        node_names = {node.get_name() for node in graph.nodes()}
        self.assertTrue(set(self.dpda.states).issubset(node_names))
        self.assertEqual(len(self.dpda.states) + 1, len(node_names))

        for state in self.dpda.states:
            node = graph.get_node(state)
            expected_shape = (
                "doublecircle" if state in self.dpda.final_states else "circle"
            )
            self.assertEqual(node.attr["shape"], expected_shape)

        expected_transitions = {
            ("q0", "a, 0 | 10", "q1"),
            ("q1", "a, 1 | 11", "q1"),
            ("q1", "b, 1 | ε", "q2"),
            ("q2", "b, 1 | ε", "q2"),
            ("q2", "ε, 0 | 0", "q3"),
        }
        seen_transitions = {
            (edge[0], edge.attr["label"], edge[1]) for edge in graph.edges()
        }
        self.assertTrue(expected_transitions.issubset(seen_transitions))
        self.assertEqual(len(expected_transitions) + 1, len(seen_transitions))

        source, symbol, dest = list(seen_transitions - expected_transitions)[0]
        self.assertEqual(symbol, "")
        self.assertEqual(dest, self.dpda.initial_state)
        self.assertTrue(source not in self.dpda.states)

    def test_show_diagram_exception(self) -> None:
        """Should raise exception"""

        self.assertRaises(
            exceptions.DiagramException,
            self.dpda.show_diagram,
            "ab",
            with_machine=False,
            with_stack=False,
        )

    def test_show_diagram_read_input_machine_only(self) -> None:
        """
        Should construct the diagram with machine only for a DPDA reading input.
        """
        input_strings = ["ab", "aabb", "aaabbb"]

        for input_string in input_strings:
            graph = self.dpda.show_diagram(input_str=input_string, with_stack=False)

            # Get edges corresponding to input path
            colored_edges = [
                edge for edge in graph.edges() if "color" in dict(edge.attr)
            ]

            edge_pairs = [
                (edge[0].state, edge[1].state)
                for edge in self.dpda._get_input_path(input_string)[0]
            ]
            self.assertEqual(edge_pairs, colored_edges)

    def test_show_diagram_read_input_stack_only(self) -> None:
        """
        Should construct the diagram with stack only for a DPDA reading input.
        """
        input_strings = ["ab", "aabb", "aaabbb"]

        for input_string in input_strings:
            graph = self.dpda.show_diagram(input_str=input_string, with_machine=False)

            # Get edges corresponding to input path
            colored_edges = [
                edge for edge in graph.edges() if "color" in dict(edge.attr)
            ]

            edge_pairs = [
                (str(edge[0]), str(edge[1]))
                for edge in self.dpda._get_input_path(input_string)[0]
            ]

            # Get stack content corresponding to input path
            nodes = [node.attr["label"] for node in graph.nodes()]

            stack_content = [
                " | " + " | ".join(reversed(edge.stack))
                for edge in list(self.dpda.read_input_stepwise(input_string))
            ]

            self.assertEqual(nodes, stack_content)
            self.assertEqual(edge_pairs, colored_edges)

    def test_show_diagram_write_file(self) -> None:
        """
        Should construct the diagram for a DPDA
        and write it to the specified file.
        """
        diagram_path = os.path.join(self.temp_dir_path, "test_dpda.png")
        try:
            os.remove(diagram_path)
        except OSError:
            pass
        self.assertFalse(os.path.exists(diagram_path))
        self.dpda.show_diagram(path=diagram_path)
        self.assertTrue(os.path.exists(diagram_path))
        os.remove(diagram_path)

    def test_show_diagram_orientations(self) -> None:
        graph = self.dpda.show_diagram()
        self.assertEqual(graph.graph_attr["rankdir"], "LR")
        graph = self.dpda.show_diagram(horizontal=False)
        self.assertEqual(graph.graph_attr["rankdir"], "TB")
        graph = self.dpda.show_diagram(reverse_orientation=True)
        self.assertEqual(graph.graph_attr["rankdir"], "RL")
        graph = self.dpda.show_diagram(horizontal=False, reverse_orientation=True)
        self.assertEqual(graph.graph_attr["rankdir"], "BT")

    def test_show_diagram_fig_size(self) -> None:
        """
        Testing figure size. Just need to make sure it matches the input
        (the library handles the rendering).
        """
        graph = self.dpda.show_diagram(fig_size=(1.1, 2))
        self.assertEqual(graph.graph_attr["size"], "1.1, 2")

        graph = self.dpda.show_diagram(fig_size=(3.3,))
        self.assertEqual(graph.graph_attr["size"], "3.3")



"""Functions for displaying and maniuplating Turing machines."""

from typing import Sequence

from automata.tm.configuration import TMConfiguration


def print_configs(validation_generator: Sequence[TMConfiguration]) -> None:
    """
    Print each machine configuration represented by the given sequence.

    Parameters
    ----------
    validation_generator : Sequence[TMConfiguration]
        A sequence containing the TMConfigurations to print.
    """

    for config in validation_generator:
        config.print()



"""Classes and methods for parsing regexes into NFAs."""

from __future__ import annotations

import copy
import re
from collections import deque
from itertools import chain, count, product, repeat
from typing import AbstractSet, Deque, Dict, Iterable, List, Optional, Set, Tuple, Type

from typing_extensions import NoReturn, Self

import automata.base.exceptions as exceptions
from automata.base.utils import get_renaming_function, pairwise
from automata.regex.lexer import Lexer, Token
from automata.regex.postfix import (
    InfixOperator,
    LeftParen,
    Literal,
    PostfixOperator,
    RightParen,
    parse_postfix_tokens,
    tokens_to_postfix,
    validate_tokens,
)

BuilderTransitionsT = Dict[int, Dict[str, Set[int]]]

RESERVED_CHARACTERS = frozenset(
    ("*", "|", "(", ")", "?", " ", "\t", "&", "+", ".", "^", "{", "}")
)


class NFARegexBuilder:
    """Builder class designed for speed in parsing regular expressions into NFAs."""

    __slots__: Tuple[str, ...] = (
        "_transitions",
        "_initial_state",
        "_final_states",
        "_state_name_counter",
    )

    _transitions: BuilderTransitionsT
    _initial_state: int
    _final_states: Set[int]
    _state_name_counter: count

    def __init__(
        self,
        *,
        transitions: BuilderTransitionsT,
        initial_state: int,
        final_states: Set[int],
        counter: count,
    ) -> None:
        """
        Initialize new builder class
        """

        self._transitions = transitions
        self._initial_state = initial_state
        self._final_states = final_states
        self._state_name_counter = counter

    @classmethod
    def from_string_literal(
        cls: Type[NFARegexBuilder], literal: str, counter: count
    ) -> NFARegexBuilder:
        """
        Initialize this builder accepting only the given string literal
        """

        transitions: BuilderTransitionsT = {
            next(counter): {symbol: set()} for symbol in literal
        }

        for start_state, path in transitions.items():
            for end_states in path.values():
                end_states.add(start_state + 1)

        final_state = next(counter)
        transitions[final_state] = {}

        return cls(
            transitions=transitions,
            initial_state=min(transitions.keys()),
            final_states={final_state},
            counter=counter,
        )

    @classmethod
    def wildcard(
        cls: Type[NFARegexBuilder], input_symbols: AbstractSet[str], counter: count
    ) -> NFARegexBuilder:
        """
        Initialize this builder for a wildcard with the given input symbols
        """

        initial_state = next(counter)
        final_state = next(counter)

        transitions: BuilderTransitionsT = {
            initial_state: {symbol: {final_state} for symbol in input_symbols},
            final_state: {},
        }

        return cls(
            transitions=transitions,
            initial_state=initial_state,
            final_states={final_state},
            counter=counter,
        )

    def union(self, other: NFARegexBuilder) -> None:
        """
        Apply the union operation to the NFA represented by this builder and other
        """
        self._transitions.update(other._transitions)

        new_initial_state = next(self._state_name_counter)

        # Add epsilon transitions from new start state to old ones
        self._transitions[new_initial_state] = {
            "": {self._initial_state, other._initial_state}
        }

        self._initial_state = new_initial_state
        self._final_states.update(other._final_states)

    def intersection(self, other: NFARegexBuilder) -> None:
        """
        Apply the intersection operation to the NFA represented by this builder
        and other. Use BFS to only traverse reachable part (keeps number of
        states down).
        """

        get_state_name = get_renaming_function(self._state_name_counter)

        new_final_states = set()
        new_transitions: BuilderTransitionsT = {}
        new_initial_state = (self._initial_state, other._initial_state)

        new_initial_state_name = get_state_name(new_initial_state)
        new_input_symbols = tuple(
            set(
                chain.from_iterable(
                    transition_dict.keys()
                    for transition_dict in chain(
                        self._transitions.values(), other._transitions.values()
                    )
                )
            )
            - {""}
        )

        queue: Deque[Tuple[int, int]] = deque()

        queue.append(new_initial_state)
        new_transitions[new_initial_state_name] = {}

        while queue:
            curr_state = queue.popleft()
            curr_state_name = get_state_name(curr_state)
            q_a, q_b = curr_state

            if q_a in self._final_states and q_b in other._final_states:
                new_final_states.add(curr_state_name)

            # States we will consider adding to the queue
            next_states_iterables: List[Iterable[Tuple[int, int]]] = []

            # Get transition dict for states in self
            transitions_a = self._transitions.get(q_a, {})
            # Add epsilon transitions for first set of transitions
            epsilon_transitions_a = transitions_a.get("")
            if epsilon_transitions_a is not None:
                state_dict = new_transitions.setdefault(curr_state_name, {})
                state_dict.setdefault("", set()).update(
                    map(get_state_name, zip(epsilon_transitions_a, repeat(q_b)))
                )
                next_states_iterables.append(zip(epsilon_transitions_a, repeat(q_b)))

            # Get transition dict for states in other
            transitions_b = other._transitions.get(q_b, {})
            # Add epsilon transitions for second set of transitions
            epsilon_transitions_b = transitions_b.get("")
            if epsilon_transitions_b is not None:
                state_dict = new_transitions.setdefault(curr_state_name, {})
                state_dict.setdefault("", set()).update(
                    map(get_state_name, zip(repeat(q_a), epsilon_transitions_b))
                )
                next_states_iterables.append(zip(repeat(q_a), epsilon_transitions_b))

            # Add all transitions moving over same input symbols
            for symbol in new_input_symbols:
                end_states_a = transitions_a.get(symbol)
                end_states_b = transitions_b.get(symbol)

                if end_states_a is not None and end_states_b is not None:
                    state_dict = new_transitions.setdefault(curr_state_name, {})
                    state_dict.setdefault(symbol, set()).update(
                        map(get_state_name, product(end_states_a, end_states_b))
                    )
                    next_states_iterables.append(product(end_states_a, end_states_b))

            # Finally, try visiting every state we found.
            for product_state in chain.from_iterable(next_states_iterables):
                product_state_name = get_state_name(product_state)
                if product_state_name not in new_transitions:
                    new_transitions[product_state_name] = {}
                    queue.append(product_state)

        self._final_states = new_final_states
        self._transitions = new_transitions
        self._initial_state = new_initial_state_name

    def concatenate(self, other: NFARegexBuilder) -> None:
        """
        Apply the concatenate operation to the NFA represented by this builder
        and other.
        """
        self._transitions.update(other._transitions)

        for state in self._final_states:
            self._transitions[state].setdefault("", set()).add(other._initial_state)

        self._final_states = other._final_states

    def repeat(self, lower_bound: int, upper_bound: Optional[int]) -> None:
        """
        Apply the repetition operator. Corresponds to repeating the NFA
        between lower_bound and upper_bound many times. If upper_bound is None,
        then the number of repetitions is unbounded.
        """
        number_of_repetitions = lower_bound if upper_bound is None else upper_bound

        prev_final_states = self._final_states

        new_initial_state = next(self._state_name_counter)
        new_transitions = copy.deepcopy(self._transitions)

        new_transitions[new_initial_state] = {"": {self._initial_state}}

        new_final_states = set()

        if lower_bound <= 1:
            new_final_states.update(self._final_states)

        # Loop around if lower bound is 0
        if lower_bound == 0:
            new_final_states.add(self._initial_state)

        prev_initial_state = self._initial_state

        for i in range(2, number_of_repetitions + 1):
            # Reset the state renaming function each time
            get_state_name = get_renaming_function(self._state_name_counter)

            # Load next copy of transitions into dict
            new_transitions.update(
                {
                    get_state_name(start_state): {
                        char: set(map(get_state_name, dest_states))
                        for char, dest_states in char_transitions.items()
                    }
                    for start_state, char_transitions in self._transitions.items()
                }
            )

            for state in prev_final_states:
                new_transitions[state].setdefault("", set()).add(
                    get_state_name(self._initial_state)
                )

            prev_final_states = set(map(get_state_name, self._final_states))
            prev_initial_state = get_state_name(self._initial_state)

            # Wonky numbering because we start with one copy of states
            if lower_bound <= i:
                new_final_states.update(prev_final_states)

        # If no upper bound, make quantifier loop around
        if upper_bound is None:
            for state in prev_final_states:
                new_transitions[state].setdefault("", set()).add(prev_initial_state)

        self._transitions = new_transitions
        self._final_states = new_final_states
        self._initial_state = new_initial_state

    def shuffle_product(self, other: NFARegexBuilder) -> None:
        """
        Apply the shuffle operation to the NFA represented by this builder and other.
        No need for BFS since all states are accessible.
        """

        get_state_name = get_renaming_function(self._state_name_counter)

        self._initial_state = get_state_name(
            (self._initial_state, other._initial_state)
        )

        new_transitions: BuilderTransitionsT = {}

        transition_product = product(
            self._transitions.items(), other._transitions.items()
        )
        for (q_a, transitions_a), (q_b, transitions_b) in transition_product:
            state_dict = new_transitions.setdefault(get_state_name((q_a, q_b)), {})

            for symbol, end_states in transitions_a.items():
                state_dict.setdefault(symbol, set()).update(
                    map(get_state_name, zip(end_states, repeat(q_b)))
                )

            for symbol, end_states in transitions_b.items():
                state_dict.setdefault(symbol, set()).update(
                    map(get_state_name, zip(repeat(q_a), end_states))
                )

        self._final_states = set(
            map(get_state_name, product(self._final_states, other._final_states))
        )
        self._transitions = new_transitions


class UnionToken(InfixOperator[NFARegexBuilder]):
    """Subclass of infix operator defining the union operator."""

    __slots__: Tuple[str, ...] = tuple()

    def get_precedence(self) -> int:
        return 1

    def op(self, left: NFARegexBuilder, right: NFARegexBuilder) -> NFARegexBuilder:
        left.union(right)
        return left


class IntersectionToken(InfixOperator[NFARegexBuilder]):
    """Subclass of infix operator defining the intersection operator."""

    __slots__: Tuple[str, ...] = tuple()

    def get_precedence(self) -> int:
        return 1

    def op(self, left: NFARegexBuilder, right: NFARegexBuilder) -> NFARegexBuilder:
        left.intersection(right)
        return left


class ShuffleToken(InfixOperator[NFARegexBuilder]):
    """Subclass of infix operator defining the shuffle operator."""

    __slots__: Tuple[str, ...] = tuple()

    def get_precedence(self) -> int:
        return 1

    def op(self, left: NFARegexBuilder, right: NFARegexBuilder) -> NFARegexBuilder:
        left.shuffle_product(right)
        return left


class KleeneStarToken(PostfixOperator[NFARegexBuilder]):
    """Subclass of postfix operator defining the kleene star operator."""

    __slots__: Tuple[str, ...] = tuple()

    def get_precedence(self) -> int:
        return 3

    def op(self, left: NFARegexBuilder) -> NFARegexBuilder:
        left.repeat(0, None)
        return left


class KleenePlusToken(PostfixOperator[NFARegexBuilder]):
    """Subclass of postfix operator defining the kleene plus operator."""

    __slots__: Tuple[str, ...] = tuple()

    def get_precedence(self) -> int:
        return 3

    def op(self, left: NFARegexBuilder) -> NFARegexBuilder:
        left.repeat(1, None)
        return left


class QuantifierToken(PostfixOperator[NFARegexBuilder]):
    """Subclass of postfix operator for repeating an expression a fixed number
    of times."""

    __slots__: Tuple[str, ...] = ("lower_bound", "upper_bound")

    def __init__(self, text: str, lower_bound: int, upper_bound: Optional[int]) -> None:
        super().__init__(text)

        if lower_bound < 0:
            raise exceptions.InvalidRegexError(
                f"Quantifier lower bound must be strictly greater than 0, not "
                f"{lower_bound}."
            )
        elif upper_bound is not None and upper_bound < lower_bound:
            raise exceptions.InvalidRegexError(
                f"Quantifier upper bound {upper_bound} inconsistent with lower "
                f"bound {lower_bound}."
            )

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    @classmethod
    def from_match(cls: Type[Self], match: re.Match) -> QuantifierToken:
        lower_bound_str = match.group(1)
        upper_bound_str = match.group(2)

        lower_bound = 0 if not lower_bound_str else int(lower_bound_str)
        upper_bound = None if not upper_bound_str else int(upper_bound_str)

        return cls(match.group(), lower_bound, upper_bound)

    def get_precedence(self) -> int:
        return 3

    def op(self, left: NFARegexBuilder) -> NFARegexBuilder:
        left.repeat(self.lower_bound, self.upper_bound)
        return left


class OptionToken(PostfixOperator[NFARegexBuilder]):
    """Subclass of postfix operator defining the option operator."""

    __slots__: Tuple[str, ...] = tuple()

    def get_precedence(self) -> int:
        return 3

    def op(self, left: NFARegexBuilder) -> NFARegexBuilder:
        left.repeat(0, 1)
        return left


class ConcatToken(InfixOperator[NFARegexBuilder]):
    """Subclass of infix operator defining the concatenation operator."""

    __slots__: Tuple[str, ...] = tuple()

    def get_precedence(self) -> int:
        return 2

    def op(self, left: NFARegexBuilder, right: NFARegexBuilder) -> NFARegexBuilder:
        left.concatenate(right)
        return left


class StringToken(Literal[NFARegexBuilder]):
    """Subclass of literal token defining a string literal."""

    __slots__: Tuple[str, ...] = ("counter",)

    def __init__(self, text: str, counter: count) -> None:
        super().__init__(text)
        self.counter = counter

    @classmethod
    def from_match(cls: Type[Self], match: re.Match) -> NoReturn:
        raise NotImplementedError

    def val(self) -> NFARegexBuilder:
        return NFARegexBuilder.from_string_literal(self.text, self.counter)


class WildcardToken(Literal[NFARegexBuilder]):
    """Subclass of literal token defining a wildcard literal."""

    __slots__: Tuple[str, ...] = ("input_symbols", "counter")

    def __init__(
        self, text: str, input_symbols: AbstractSet[str], counter: count
    ) -> None:
        super().__init__(text)
        self.input_symbols = input_symbols
        self.counter = counter

    @classmethod
    def from_match(cls: Type[Self], match: re.Match) -> NoReturn:
        raise NotImplementedError

    def val(self) -> NFARegexBuilder:
        return NFARegexBuilder.wildcard(self.input_symbols, self.counter)


def add_concat_and_empty_string_tokens(
    token_list: List[Token[NFARegexBuilder]],
    state_name_counter: count,
) -> List[Token[NFARegexBuilder]]:
    """Add concat tokens to list of parsed infix tokens."""

    final_token_list = []

    # Pairs of token types to insert concat tokens in between
    concat_pairs = [
        (Literal, Literal),
        (RightParen, LeftParen),
        (RightParen, Literal),
        (Literal, LeftParen),
        (PostfixOperator, Literal),
        (PostfixOperator, LeftParen),
    ]

    # Pairs of tokens to insert empty string literals between
    empty_string_pairs = [(LeftParen, RightParen)]

    for curr_token, next_token in pairwise(token_list, True):
        final_token_list.append(curr_token)

        if next_token is not None:
            for firstClass, secondClass in concat_pairs:
                if isinstance(curr_token, firstClass) and isinstance(
                    next_token, secondClass
                ):
                    final_token_list.append(ConcatToken(""))

            for firstClass, secondClass in empty_string_pairs:
                if isinstance(curr_token, firstClass) and isinstance(
                    next_token, secondClass
                ):
                    final_token_list.append(StringToken("", state_name_counter))

    return final_token_list


def get_regex_lexer(
    input_symbols: AbstractSet[str], state_name_counter: count
) -> Lexer[NFARegexBuilder]:
    """Get lexer for parsing regular expressions."""
    lexer: Lexer[NFARegexBuilder] = Lexer()

    lexer.register_token(LeftParen.from_match, r"\(")
    lexer.register_token(RightParen.from_match, r"\)")
    lexer.register_token(UnionToken.from_match, r"\|")
    lexer.register_token(IntersectionToken.from_match, r"\&")
    lexer.register_token(ShuffleToken.from_match, r"\^")
    lexer.register_token(KleeneStarToken.from_match, r"\*")
    lexer.register_token(KleenePlusToken.from_match, r"\+")
    lexer.register_token(OptionToken.from_match, r"\?")
    lexer.register_token(QuantifierToken.from_match, r"\{(.*?),(.*?)\}")
    lexer.register_token(
        lambda match: WildcardToken(match.group(), input_symbols, state_name_counter),
        r"\.",
    )
    lexer.register_token(
        lambda match: StringToken(match.group(), state_name_counter), r"\S"
    )

    return lexer


def parse_regex(regexstr: str, input_symbols: AbstractSet[str]) -> NFARegexBuilder:
    """Return an NFARegexBuilder corresponding to regexstr."""

    if len(regexstr) == 0:
        return NFARegexBuilder.from_string_literal(regexstr, count(0))

    state_name_counter = count(0)

    lexer = get_regex_lexer(input_symbols, state_name_counter)
    lexed_tokens = lexer.lex(regexstr)
    validate_tokens(lexed_tokens)
    tokens_with_concats = add_concat_and_empty_string_tokens(
        lexed_tokens, state_name_counter
    )
    postfix = tokens_to_postfix(tokens_with_concats)

    return parse_postfix_tokens(postfix)



"""Classes and methods for working with all pushdown automata."""

from __future__ import annotations

import abc
import os
from collections import defaultdict
from typing import (
    AbstractSet,
    Any,
    DefaultDict,
    Generator,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
)

import automata.base.exceptions as exceptions
import automata.pda.exceptions as pda_exceptions
from automata.base.automaton import Automaton, AutomatonStateT, AutomatonTransitionsT
from automata.base.utils import (
    LayoutMethod,
    create_graph,
    create_unique_random_id,
    save_graph,
)
from automata.pda.configuration import PDAConfiguration
from automata.pda.stack import PDAStack

# Optional imports for use with visual functionality
try:
    import coloraide
    import pygraphviz as pgv
except ImportError:
    _visual_imports = False
else:
    _visual_imports = True

PDAStateT = AutomatonStateT
PDATransitionsT = AutomatonTransitionsT
PDAAcceptanceModeT = Literal["final_state", "empty_stack", "both"]
EdgeDrawnDictT = DefaultDict[Tuple[Any, Any, Tuple[str, str, str]], bool]


class PDA(Automaton, metaclass=abc.ABCMeta):
    """An abstract base class for pushdown automata."""

    __slots__ = tuple()

    stack_symbols: AbstractSet[str]
    initial_stack_symbol: str
    acceptance_mode: PDAAcceptanceModeT

    @staticmethod
    def _get_edge_name(
        input_symbol: str = "", stack_top_symbol: str = "", stack_push: str = ""
    ) -> str:
        return (
            ("ε" if input_symbol == "" else str(input_symbol))
            + ", "
            + ("ε" if stack_top_symbol == "" else str(stack_top_symbol))
            + " | "
            + ("ε" if stack_push == "" else str(stack_push))
        )

    @staticmethod
    def _get_symbol_configuration(
        from_state: PDAConfiguration, to_state: PDAConfiguration
    ) -> Tuple[str, str, str]:
        if (
            from_state.remaining_input == to_state.remaining_input
            or len(from_state.remaining_input) == 0
        ):
            input_symbol = ""
        else:
            input_symbol = from_state.remaining_input[0]

        stack_top_symbol = from_state.stack.top()

        if len(from_state.stack) == len(to_state.stack) + 1:
            stack_push_symbols = ""
        else:
            stack_push_symbols = "".join(to_state.stack[len(from_state.stack) - 1 :])

        return (input_symbol, stack_top_symbol, stack_push_symbols[::-1])

    @abc.abstractmethod
    def iter_transitions(
        self,
    ) -> Generator[Tuple[PDAStateT, PDAStateT, Tuple[str, str, str]], None, None]:
        """
        Iterate over all transitions in the automaton.
        Each transition is a tuple of the form
        (from_state, to_state, (input_symbol, stack_top_symbol, stack_push_symbols))
        """

        raise NotImplementedError(
            f"iter_transitions is not implemented for {self.__class__}"
        )

    def show_diagram(
        self,
        input_str: Optional[str] = None,
        with_machine: bool = True,
        with_stack: bool = True,
        path: Union[str, os.PathLike, None] = None,
        *,
        layout_method: LayoutMethod = "dot",
        horizontal: bool = True,
        reverse_orientation: bool = False,
        fig_size: Union[Tuple[float, float], Tuple[float], None] = None,
        font_size: float = 14.0,
        arrow_size: float = 0.85,
        state_separation: float = 0.5,
    ) -> pgv.AGraph:
        """
        Generates the graph associated with the given PDA.
        Args:
            - input_str (str, optional): String list of input symbols. Defaults to None.
            - with_machine (bool, optional): Constructs the diagram with states and
              transitions. Ignored if `input_str` is None. Default to True.
            - with_stack (bool, optional): Constructs the diagram with stack and its
              operations. Ignored if `input_str` is None. Default to True.
            - path (str or os.PathLike, optional): Path to output file. If
              None, the output will not be saved.
            - horizontal (bool, optional): Direction of node layout. Defaults
              to True.
            - reverse_orientation (bool, optional): Reverse direction of node
              layout. Defaults to False.
            - fig_size (tuple, optional): Figure size. Defaults to None.
            - font_size (float, optional): Font size. Defaults to 14.0.
            - arrow_size (float, optional): Arrow head size. Defaults to 0.85.
            - state_separation (float, optional): Node distance. Defaults to 0.5.
        Returns:
            AGraph corresponding to the given automaton.
        """

        if not _visual_imports:
            raise ImportError(
                "Missing visualization packages; "
                "please install coloraide and pygraphviz."
            )

        # Defining the graph.
        graph = create_graph(
            horizontal, reverse_orientation, fig_size, state_separation
        )

        font_size_str = str(font_size)
        arrow_size_str = str(arrow_size)

        is_edge_drawn: EdgeDrawnDictT = defaultdict(lambda: False)
        if input_str is not None:
            if not (with_machine or with_stack):
                raise exceptions.DiagramException(
                    "Both with_machine and with_stack cannot be False."
                    " This will produce a empty diagram"
                )

            input_path, is_accepted = self._get_input_path(input_str)

            start_color = coloraide.Color("#ff0")
            end_color = (
                coloraide.Color("#0f0") if is_accepted else coloraide.Color("#f00")
            )

            if with_machine:
                # initialize diagram with all states
                self._add_states_diagram(graph, arrow_size_str, font_size_str)

                # add required transitions to show execution of the
                # PDA for the given input string
                self._create_transitions_for_input_diagram(
                    graph,
                    input_path,
                    is_edge_drawn,
                    start_color,
                    end_color,
                    arrow_size_str,
                    font_size_str,
                )

                # add all the necessary transitions between states
                self._add_transitions_diagram(
                    graph, is_edge_drawn, arrow_size_str, font_size_str
                )

            if with_stack:
                # add the stack transitions
                self._create_stack_diagram(
                    input_path,
                    graph,
                    start_color,
                    end_color,
                    font_size_str,
                    arrow_size_str,
                )
        else:
            # initialize diagram with all states
            self._add_states_diagram(graph, arrow_size_str, font_size_str)

            # add all the necessary transitions between states
            self._add_transitions_diagram(
                graph, is_edge_drawn, arrow_size_str, font_size_str
            )

        # Set layout
        graph.layout(prog=layout_method)

        # Write diagram to file
        if path is not None:
            save_graph(graph, path)

        return graph

    def _create_stack_diagram(
        self,
        input_path: List[Tuple[PDAConfiguration, PDAConfiguration]],
        graph: pgv.AGraph,
        start_color: coloraide.Color,
        end_color: coloraide.Color,
        font_size: str,
        arrow_size: str,
    ) -> None:
        """
        Constructs stack for all the transitions in the `input_path` and
        adds the constructed stacks into `graph`. Returns the same `graph`
        """
        from_node = input_path[0][0]
        label = " | ".join(reversed(from_node.stack))
        graph.add_node(
            from_node,
            label=f" | {label}",
            shape="record",
            arrowsize=arrow_size,
            fontsize=font_size,
        )

        interpolation = coloraide.Color.interpolate(
            [start_color, end_color], space="srgb"
        )

        for i, (c, n) in enumerate(input_path, start=1):
            from_node = n
            label = " | ".join(reversed(from_node.stack))

            color = interpolation(i / len(input_path))

            graph.add_node(
                from_node,
                label=f" | {label}",
                shape="record",
                arrowsize=arrow_size,
                fontsize=font_size,
            )
            graph.add_edge(
                c,
                n,
                label=f"<<b>[<i>#{i}</i>]</b>>",
                arrowsize=arrow_size,
                fontsize=font_size,
                color=color.to_string(hex=True),
                penwidth="2.5",
            )

        return graph

    def _create_transitions_for_input_diagram(
        self,
        graph: pgv.AGraph,
        input_path: List[Tuple[PDAConfiguration, PDAConfiguration]],
        edge_drawn: EdgeDrawnDictT,
        start_color: coloraide.Color,
        end_color: coloraide.Color,
        arrow_size: str,
        font_size: str,
    ) -> None:
        """
        Add transitions to show execution of the PDA for the given input string
        """

        interpolation = coloraide.Color.interpolate(
            [start_color, end_color], space="srgb"
        )

        # find all transitions in the finite state machine with traversal.
        for transition_index, (from_state, to_state) in enumerate(input_path, start=1):
            color = interpolation(transition_index / len(input_path))

            symbol = self._get_symbol_configuration(from_state, to_state)
            label = self._get_edge_name(*symbol)

            edge_drawn[from_state.state, to_state.state, symbol] = True
            graph.add_edge(
                self._get_state_name(from_state.state),
                self._get_state_name(to_state.state),
                label=f"<{label} <b>[<i>#{transition_index}</i>]</b>>",
                arrowsize=arrow_size,
                fontsize=font_size,
                color=color.to_string(hex=True),
                penwidth="2.5",
            )

    def _add_transitions_diagram(
        self,
        graph: pgv.AGraph,
        is_edge_drawn: EdgeDrawnDictT,
        arrow_size: str,
        font_size: str,
    ) -> None:
        """
        Add transitions to between states
        """

        edge_labels = defaultdict(list)
        for from_state, to_state, symbol in self.iter_transitions():
            if is_edge_drawn[from_state, to_state, symbol]:
                continue

            from_node = self._get_state_name(from_state)
            to_node = self._get_state_name(to_state)
            label = self._get_edge_name(*symbol)
            edge_labels[from_node, to_node].append(label)

        for (from_node, to_node), labels in edge_labels.items():
            graph.add_edge(
                from_node,
                to_node,
                label="\n".join(sorted(labels)),
                arrowsize=arrow_size,
                fontsize=font_size,
            )

    def _add_states_diagram(
        self,
        graph: pgv.AGraph,
        arrow_size: str,
        font_size: str,
    ) -> None:
        """
        Add all the states of the PDA
        """

        # create unique id to avoid colliding with other states
        null_node = create_unique_random_id()

        graph.add_node(
            null_node,
            label="",
            tooltip=".",
            shape="point",
            fontsize=font_size,
        )
        initial_node = self._get_state_name(self.initial_state)
        graph.add_edge(
            null_node,
            initial_node,
            tooltip="->" + initial_node,
            arrowsize=arrow_size,
        )

        nonfinal_states = map(self._get_state_name, self.states - self.final_states)
        final_states = map(self._get_state_name, self.final_states)
        graph.add_nodes_from(nonfinal_states, shape="circle", fontsize=font_size)
        graph.add_nodes_from(final_states, shape="doublecircle", fontsize=font_size)

    def _validate_transition_invalid_input_symbols(
        self, start_state: PDAStateT, input_symbol: str
    ) -> None:
        """Raise an error if transition input symbols are invalid."""
        if input_symbol not in self.input_symbols and input_symbol != "":
            raise exceptions.InvalidSymbolError(
                "state {} has invalid transition input symbol {}".format(
                    start_state, input_symbol
                )
            )

    def _validate_transition_invalid_stack_symbols(
        self, start_state: PDAStateT, stack_symbol: str
    ) -> None:
        """Raise an error if transition stack symbols are invalid."""
        if stack_symbol not in self.stack_symbols:
            raise exceptions.InvalidSymbolError(
                "state {} has invalid transition stack symbol {}".format(
                    start_state, stack_symbol
                )
            )

    def _validate_initial_stack_symbol(self) -> None:
        """Raise an error if initial stack symbol is invalid."""
        if self.initial_stack_symbol not in self.stack_symbols:
            raise exceptions.InvalidSymbolError(
                "initial stack symbol {} is invalid".format(self.initial_stack_symbol)
            )

    def _validate_acceptance(self) -> None:
        """Raise an error if the acceptance mode is invalid."""
        if self.acceptance_mode not in ("final_state", "empty_stack", "both"):
            raise pda_exceptions.InvalidAcceptanceModeError(
                "acceptance mode {} is invalid".format(self.acceptance_mode)
            )

    @abc.abstractmethod
    def _validate_transition_invalid_symbols(
        self, start_state: PDAStateT, paths: PDATransitionsT
    ) -> None:
        pass

    @abc.abstractmethod
    def _get_input_path(
        self, input_str: str
    ) -> Tuple[List[Tuple[PDAConfiguration, PDAConfiguration]], bool]:
        """Calculate the path taken by input."""

        raise NotImplementedError(
            f"_get_input_path is not implemented for {self.__class__}"
        )

    def validate(self) -> None:
        """
        Raises an exception if this automaton is not internally consistent.

        Raises
        ------
        InvalidStateError
            If this PDA has invalid states in the transition dictionary.
        MissingStateError
            If this PDA has states missing from the transition dictionary.
        InvalidSymbolError
            If this PDA has invalid symbols in the transition dictionary.
        InvalidAcceptanceModeError
            If this PDA has an invalid acceptance mode.
        """
        for start_state, paths in self.transitions.items():
            self._validate_transition_invalid_symbols(start_state, paths)
        self._validate_initial_state()
        self._validate_initial_stack_symbol()
        self._validate_final_states()
        self._validate_acceptance()

    def _has_lambda_transition(self, state: PDAStateT, stack_symbol: str) -> bool:
        """Return True if the current config has any lambda transitions."""
        return (
            state in self.transitions
            and "" in self.transitions[state]
            and stack_symbol in self.transitions[state][""]
        )

    def _replace_stack_top(self, stack: PDAStack, new_stack_top: str) -> PDAStack:
        """Replace the top of the PDA stack with another symbol"""
        if new_stack_top == "":
            new_stack = stack.pop()
        else:
            new_stack = stack.replace(new_stack_top)
        return new_stack

    def _has_accepted(self, current_configuration: PDAConfiguration) -> bool:
        """Check whether the given config indicates accepted input."""
        # If there's input left, we're not finished.
        if current_configuration.remaining_input:
            return False
        if self.acceptance_mode in ("empty_stack", "both"):
            # If the stack is empty, we accept.
            if not current_configuration.stack:
                return True
        if self.acceptance_mode in ("final_state", "both"):
            # If current state is a final state, we accept.
            if current_configuration.state in self.final_states:
                return True
        # Otherwise, not.
        return False



"""Classes and functions for testing the behavior of PDAConfigurations."""

import tests.test_pda as test_pda
from automata.pda.configuration import PDAConfiguration
from automata.pda.stack import PDAStack


class TestPDAConfiguration(test_pda.TestPDA):
    """A test class for testing configurations of pushdown automata."""

    def test_config_hashability(self) -> None:
        self.assertEqual(
            hash(PDAConfiguration("q0", "ab", PDAStack(["a", "b"]))),
            hash(PDAConfiguration("q0", "ab", PDAStack(["a", "b"]))),
        )

    def test_config_repr(self) -> None:
        """Should create proper string representation of PDA configuration."""
        config = PDAConfiguration("q0", "ab", PDAStack(["a", "b"]))
        self.assertEqual(
            repr(config),
            "PDAConfiguration('q0', 'ab', PDAStack(('a', 'b')))",  # noqa
        )

    def test_config_equality(self) -> None:
        """Should only be equal for equal configurations."""
        config = PDAConfiguration("q0", "ab", PDAStack(["a", "b"]))

        self.assertEqual(config, config)
        self.assertEqual(config, PDAConfiguration("q0", "ab", PDAStack(["a", "b"])))

        self.assertNotEqual(config, "")
        self.assertNotEqual(config, PDAConfiguration("q1", "ab", PDAStack(["a", "b"])))
        self.assertNotEqual(config, PDAConfiguration("q0", "ab", PDAStack(["b", "b"])))



"""Tests of the postfix conversion and parsing utility functions."""

import unittest
from typing import List, TypeVar

import automata.base.exceptions as exceptions
import automata.regex.postfix as postfix
from automata.regex.lexer import Lexer, Token

ResultT = TypeVar("ResultT")


class Integer(postfix.Literal[int]):
    def val(self) -> int:
        """It evaluates to its (integer) value."""
        return int(self.text)


class Add(postfix.InfixOperator[int]):
    """Addition."""

    def get_precedence(self) -> int:
        return 10  # Precedence: higher than integers, lower than mult

    def op(self, left: int, right: int) -> int:
        return left + right


class Minus(postfix.InfixOperator[int]):
    """Subtraction."""

    def get_precedence(self) -> int:
        return 10  # Precedence: higher than integers, lower than mult

    def op(self, left: int, right: int) -> int:
        return left - right


class Mult(postfix.InfixOperator[int]):
    """Multiplication."""

    def get_precedence(self) -> int:
        return 20  # Higher precedence than addition/substraction.

    def op(self, left: int, right: int) -> int:
        return left * right


class Divide(postfix.InfixOperator[int]):
    """Division."""

    def get_precedence(self) -> int:
        return 20  # Same precedence than multiplication

    def op(self, left: int, right: int) -> int:
        return left // right


class TestTokens(unittest.TestCase):
    """Test token subclasses for not implemented errors."""

    def test_token_abstract_methods_not_implemented(self) -> None:
        """Should raise NotImplementedError when calling abstract methods."""

        with self.assertRaises(NotImplementedError):
            getattr(postfix.Operator, "get_precedence")(postfix.Operator)

        with self.assertRaises(NotImplementedError):
            getattr(postfix.InfixOperator, "op")(postfix.InfixOperator, None, None)

        with self.assertRaises(NotImplementedError):
            getattr(postfix.PostfixOperator, "op")(postfix.PostfixOperator, None)

        with self.assertRaises(NotImplementedError):
            getattr(postfix.Literal, "val")(postfix.Literal)


class TestArithmeticParser(unittest.TestCase):
    """Test parsing arithmetic expressions."""

    arithmetic_lexer: Lexer

    def test_parse_invalid_token(self) -> None:
        """Test exception for invalid input tokens."""
        with self.assertRaises(exceptions.InvalidRegexError):
            postfix.parse_postfix_tokens([""])  # type: ignore

    def test_nested_parenthesized_expression(self) -> None:
        """Test parsing parenthesized expression."""
        # Parsing:
        # "( 4 + ( 1 + 2 * 3 * ( 4 + 5 ) + 6 ) ) * 7 + 8"
        tokens = [
            postfix.LeftParen("("),
            Integer("4"),
            Add("+"),
            postfix.LeftParen("("),
            Integer("1"),
            Add("+"),
            Integer("2"),
            Mult("*"),
            Integer("3"),
            Mult("*"),
            postfix.LeftParen("("),
            Integer("4"),
            Add("+"),
            Integer("5"),
            postfix.RightParen(")"),
            Add("+"),
            Integer("6"),
            postfix.RightParen(")"),
            postfix.RightParen(")"),
            Mult("*"),
            Integer("7"),
            Add("+"),
            Integer("8"),
        ]

        postfix_tokens = postfix.tokens_to_postfix(tokens)
        res = postfix.parse_postfix_tokens(postfix_tokens)
        self.assertEqual((4 + (1 + 2 * 3 * (4 + 5) + 6)) * 7 + 8, res)

    def setUp(self) -> None:
        self.arithmetic_lexer: Lexer = Lexer()

        self.arithmetic_lexer.register_token(postfix.LeftParen.from_match, r"\(")
        self.arithmetic_lexer.register_token(postfix.RightParen.from_match, r"\)")
        self.arithmetic_lexer.register_token(Integer.from_match, r"[0-9]+")
        self.arithmetic_lexer.register_token(Add.from_match, r"\+")
        self.arithmetic_lexer.register_token(Minus.from_match, r"-")
        self.arithmetic_lexer.register_token(Mult.from_match, r"\*")
        self.arithmetic_lexer.register_token(Divide.from_match, r"/")

    def test_expression_invalid_ordering(self) -> None:
        """Check for exception raised when lexing invalid regular expressions."""

        with self.assertRaises(exceptions.InvalidRegexError):
            postfix.validate_tokens(self.arithmetic_lexer.lex("+6"))

        with self.assertRaises(exceptions.InvalidRegexError):
            postfix.validate_tokens(self.arithmetic_lexer.lex("+5+"))

        with self.assertRaises(exceptions.InvalidRegexError):
            postfix.validate_tokens(self.arithmetic_lexer.lex("6/"))

        with self.assertRaises(exceptions.InvalidRegexError):
            postfix.validate_tokens(self.arithmetic_lexer.lex("1 + 2 - + 3"))

        with self.assertRaises(exceptions.InvalidRegexError):
            postfix.validate_tokens(self.arithmetic_lexer.lex(")("))

        with self.assertRaises(exceptions.InvalidRegexError):
            postfix.validate_tokens(self.arithmetic_lexer.lex("(((2))"))

        with self.assertRaises(exceptions.InvalidRegexError):
            postfix.validate_tokens(self.arithmetic_lexer.lex("(+5"))

        with self.assertRaises(exceptions.InvalidRegexError):
            postfix.validate_tokens(self.arithmetic_lexer.lex("6/)"))

    def parse(self, tokens: List[Token[ResultT]]) -> ResultT:
        """Helper function for parsing token list tokens"""
        postfix_tokens = postfix.tokens_to_postfix(tokens)
        return postfix.parse_postfix_tokens(postfix_tokens)

    def test_single_number(self) -> None:
        """Test parsing a single number."""
        val = self.parse(self.arithmetic_lexer.lex("13"))
        self.assertEqual(13, val)

    def test_negative_number(self) -> None:
        """Test parsing a negative number."""
        val = self.parse(self.arithmetic_lexer.lex("0-13"))
        self.assertEqual(-13, val)

    def test_simple_mult(self) -> None:
        """Test parsing simple multiplication."""
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("2 * 4")))
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("2 * 2 * 2")))
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("2 * (2 * 2)")))
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("(2 * 2) * 2")))
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("(2 + 2) * 2")))

    def test_precedence(self) -> None:
        """Test checking for correct precedence."""
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("2 * 3 + 2")))
        self.assertEqual(10, self.parse(self.arithmetic_lexer.lex("2 * (3+2)")))

    def test_negative_mult(self) -> None:
        """Test multiplying negative numbers."""
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("(0-2) * (0- 4)")))
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("(0-2) * 2 * (0-2)")))
        self.assertEqual(-5, self.parse(self.arithmetic_lexer.lex("1 + (0-2) * 3")))

    def test_division(self) -> None:
        """Test dividing numbers."""
        self.assertEqual(2, self.parse(self.arithmetic_lexer.lex("4 / 2")))
        self.assertEqual(2, self.parse(self.arithmetic_lexer.lex("5 / 2")))
        self.assertEqual(2, self.parse(self.arithmetic_lexer.lex("6 - 8 / 2")))
        self.assertEqual(3, self.parse(self.arithmetic_lexer.lex("3 * 2 / 2")))
        self.assertEqual(3, self.parse(self.arithmetic_lexer.lex("2 * 3 / 2")))
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("8 / 2 * 2")))
        self.assertEqual(8, self.parse(self.arithmetic_lexer.lex("(8 / 2) * 2")))
        self.assertEqual(2, self.parse(self.arithmetic_lexer.lex("8 / (2 * 2)")))
        self.assertEqual(2, self.parse(self.arithmetic_lexer.lex("16/4/2")))



"""Classes and functions for testing the behavior of Regex tools"""

import re as regex
import unittest

import automata.base.exceptions as exceptions
import automata.regex.regex as re
from automata.fa.nfa import NFA
from automata.regex.parser import StringToken, WildcardToken


class TestRegex(unittest.TestCase):
    """A test class for testing regular expression tools"""

    def test_validate_valid(self) -> None:
        """Should pass validation for valid regular expression"""
        re.validate("a*")
        re.validate("b|a?*")

    def test_validate_invalid(self) -> None:
        """Should raise error for invalid regular expressions"""
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "ab|")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "?")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "a|b|*")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "a||b")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "((abc*)))((abd)")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "*")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "ab(bc)*((bbcd)")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "a(*)")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "a(|)")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "a{1,0}")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "a{-1,}")
        self.assertRaises(exceptions.InvalidRegexError, re.validate, "a{-2,-1}")

    def test_invalid_token_creation(self) -> None:
        """Should raise error for invalid class creation"""
        match_obj = regex.compile("a").match("a")
        self.assertRaises(NotImplementedError, StringToken.from_match, match_obj)
        self.assertRaises(NotImplementedError, WildcardToken.from_match, match_obj)

    def test_helper_validate_invalid(self) -> None:
        """Should pass validation for valid regular expression"""
        self.assertFalse(re._validate("a(|)"))

    def test_validate_unicode_characters(self) -> None:
        """Should pass validation for regular expressions with unicode characters."""
        re.validate("(µ|🤖ù)*")

    def test_unicode_input_symbols(self) -> None:
        """Should have correct unicode input symbols."""
        nfa = NFA.from_regex("(µ🔥|🔥✨?)*")
        self.assertEqual(nfa.input_symbols, {"µ", "🔥", "✨"})

    def test_isequal(self) -> None:
        """Should correctly check equivalence of two regular expressions"""

        self.assertTrue(re.isequal("aa?", "a|aa"))
        self.assertTrue(re.isequal("a(a*b|b)", "aaa*b|ab"))
        self.assertTrue(
            re.isequal(
                "a(a*b|b)b(cd*|dc*)",
                "(aaa*bbcd|abbcd)d*|(aaa*bb(dcc*|(d|c))|abb(dcc*|(d|c)))",
            )
        )
        self.assertTrue(
            re.isequal(
                "(aaa*bbcd|abbcd)d*|(aaa*bb(dcc*|(d|c))|abb(dcc*|(d|c)))",
                "((aaaa*bbcd|aabbcd)d|abbcdd)d*|((aaaa*bb|aabb)dccc*|"
                "((aaaa*bbcd|aabbcd)|((aaaa*bb|aabb)(dc|(c|d))|(abbdccc*|(abb(dc|(c|d))|abbcd)))))",  # noqa
            )
        )

    def test_not_isequal(self) -> None:
        """Should correctly check non-equivalence of two regular expressions"""

        self.assertFalse(
            re.isequal(
                "baaa*b(b|a)|(bab(b|a)|(bb|ba))",
                "baaaa*b(a|b)|(baab(a|b)|bab(bb|(a|(b|ba))))",
            )
        )

    def test_issubset(self) -> None:
        """Should correctly verify if re1 is subset of re2"""

        self.assertTrue(re.issubset("aa?", "a*"))
        self.assertFalse(re.issubset("a*", "a?"))
        self.assertTrue(re.issubset("aaa*b|bc", "a*b|b*c*"))

    def test_issuperset(self) -> None:
        """Should correctly verify if re1 is superset of re2"""

        self.assertFalse(re.issuperset("aa?", "a*"))
        self.assertTrue(re.issuperset("a*", "a?"))
        self.assertTrue(re.issuperset("a*b|b*c*", "aaa*b|bc"))

    def test_intersection(self) -> None:
        """Should correctly check intersection of two regular expressions"""
        # Basic test
        nfa_1 = NFA.from_regex("(0|(01))&(01)")
        nfa_2 = NFA.from_regex("01")

        self.assertEqual(nfa_1, nfa_2)

        # Test intersection with NFA function on unrelated regexes
        regex_1 = "a|abacd"
        regex_2 = "a(a*b|b)b(cd*|dc*)"
        nfa_3 = NFA.from_regex(regex_1).intersection(NFA.from_regex(regex_2))
        nfa_4 = NFA.from_regex(f"({regex_1})&({regex_2})")

        self.assertEqual(nfa_3, nfa_4)

        # Test intersection subset
        regex_3 = "bcdaaa"
        nfa_5 = NFA.from_regex(regex_3)
        nfa_6 = NFA.from_regex(f"({regex_3}) & (bcda*)")

        self.assertEqual(nfa_5, nfa_6)

        # Test distributive law
        regex_4 = f"{regex_1} & (({regex_2}) | ({regex_3}))"
        regex_5 = f"(({regex_1}) & ({regex_2})) | (({regex_1}) & ({regex_3}))"
        nfa_7 = NFA.from_regex(regex_4)
        nfa_8 = NFA.from_regex(regex_5)

        self.assertEqual(nfa_7, nfa_8)

    def test_kleene_plus(self) -> None:
        """Should correctly check kleene plus of two regular expressions"""
        # Basic test
        self.assertTrue(re.isequal("aa*", "a+"))
        self.assertTrue(re.isequal("(abc)(abc)*", "(abc)+"))
        self.assertTrue(re.isequal("a&a+", "a"))

        self.assertFalse(re.isequal("a*", "a+"))
        self.assertTrue(re.issuperset("a*", "a+"))

    def test_wildcard(self) -> None:
        """Should correctly check wildcard"""

        input_symbols = {"a", "b", "c"}

        self.assertTrue(re.isequal("a|b|c", ".", input_symbols=input_symbols))
        self.assertTrue(
            re.isequal("(abc)|(aac)|(acc)", "a.c", input_symbols=input_symbols)
        )
        self.assertTrue(re.isequal("a&.", "a", input_symbols=input_symbols))

        self.assertTrue(re.issubset("a.b", "...", input_symbols=input_symbols))
        self.assertTrue(re.issuperset(".", "a|b", input_symbols=input_symbols))

    def test_shuffle(self) -> None:
        """Should correctly check shuffle"""

        input_symbols = {"a", "b", "c", "d"}

        self.assertTrue(re.isequal("a^b", "ab|ba", input_symbols=input_symbols))
        self.assertTrue(
            re.isequal(
                "ab^cd",
                "abcd | acbd | cabd | acdb | cadb | cdab",
                input_symbols=input_symbols,
            )
        )
        self.assertTrue(
            re.isequal("(a*)^(b*)^(c*)^(d*)", ".*", input_symbols=input_symbols)
        )
        self.assertTrue(
            re.isequal("ca^db", "(c^db)a | (ca^d)b", input_symbols=input_symbols)
        )
        self.assertTrue(
            re.isequal("a^(b|c)", "ab | ac | ba | ca", input_symbols=input_symbols)
        )

        reference_nfa = NFA.from_regex("a*^ba")
        other_nfa = NFA.shuffle_product(NFA.from_regex("a*"), NFA.from_regex("ba"))
        self.assertEqual(reference_nfa, other_nfa)

    def test_quantifier(self) -> None:
        """Should correctly check quantifier"""

        input_symbols = {"a", "b", "c", "d"}

        # Simple equivalences
        self.assertTrue(re.isequal("a{1,3}", "a|aa|aaa", input_symbols=input_symbols))
        self.assertTrue(re.isequal("a{5,5}", "aaaaa", input_symbols=input_symbols))
        self.assertTrue(re.isequal("a{1,}", "a+", input_symbols=input_symbols))
        self.assertTrue(re.isequal("a{0,}", "a*", input_symbols=input_symbols))
        self.assertTrue(re.isequal("a{4,}", "aaaa+", input_symbols=input_symbols))
        self.assertTrue(
            re.isequal("a{,4}", "a?|aa|aaa|aaaa", input_symbols=input_symbols)
        )

        # More complex equivalences
        self.assertTrue(re.isequal("ba{,1}", "ba?", input_symbols=input_symbols))
        self.assertTrue(
            re.isequal("(b|a){0,2}", "(a?)|b|ab|ba|bb|aa", input_symbols=input_symbols)
        )
        self.assertTrue(
            re.isequal("(a*b|b*c*){0,1}", "(a*b|b*c*)?", input_symbols=input_symbols)
        )
        self.assertTrue(
            re.isequal(
                "(aa^bb|ca^cb){0,}", "(aa^bb|ca^cb)*", input_symbols=input_symbols
            )
        )
        self.assertTrue(
            re.isequal(
                "(aa|bb^ca|cb){1,}", "(aa|bb^ca|cb)+", input_symbols=input_symbols
            )
        )

        # Tests for multiple quantifiers
        self.assertTrue(
            re.isequal("a{1,2}b{1,2}", "ab|aab|abb|aabb", input_symbols=input_symbols)
        )
        self.assertTrue(
            re.isequal("a{2,2}(c*b){3,3}", "aac*bc*bc*b", input_symbols=input_symbols)
        )
        self.assertTrue(
            re.isequal("a{2,2}ccb{3,3}", "aaccbbb", input_symbols=input_symbols)
        )

    def test_blank(self) -> None:
        """Should correctly parse blank"""
        self.assertTrue(re.isequal("()", ""))
        self.assertTrue(re.isequal("a|()", "a?"))
        self.assertTrue(re.isequal("a()", "a"))
        self.assertTrue(re.isequal("a()b()()c()", "abc"))

    def test_invalid_symbols(self) -> None:
        """Should throw exception if reserved character is in input symbols"""
        with self.assertRaises(exceptions.InvalidSymbolError):
            NFA.from_regex("a+", input_symbols={"a", "+"})



"""Classes and functions for testing the behavior of PDAStacks."""

import tests.test_pda as test_pda
from automata.pda.stack import PDAStack


class TestPDAStack(test_pda.TestPDA):
    """A test class for testing stacks of pushdown automata."""

    stack: PDAStack

    def setUp(self) -> None:
        self.stack = PDAStack(["a", "b"])

    def test_stack_hashability(self) -> None:
        self.assertEqual(hash(self.stack), hash(PDAStack(["a", "b"])))

    def test_stack_iter(self) -> None:
        """Should loop through the PDA stack in some manner."""
        self.assertEqual(list(self.stack), ["a", "b"])

    def test_stack_get(self) -> None:
        """Should retrieve indices in the PDA stack in some manner."""
        self.assertEqual(self.stack[0], "a")
        self.assertEqual(self.stack[1], "b")

    def test_stack_repr(self) -> None:
        """Should create proper string representation of PDA stack."""
        self.assertEqual(repr(self.stack), "PDAStack(('a', 'b'))")

    def test_stack_equality(self) -> None:
        """Should only be equal for equal configurations."""
        stack = PDAStack(["a", "b"])

        self.assertEqual(stack, stack)
        self.assertEqual(stack, PDAStack(["a", "b"]))

        self.assertNotEqual(stack, "")
        self.assertNotEqual(stack, PDAStack(["a", "a"]))
        self.assertNotEqual(stack, PDAStack(["b", "b"]))



"""Classes and methods for converting lists of tokens to postfix ordering."""

import abc
from collections import deque
from itertools import chain, zip_longest
from typing import Deque, Iterable, List, Optional, Tuple, TypeVar, cast

import automata.base.exceptions as exceptions
from automata.regex.lexer import Token

ExpressionResultT = TypeVar("ExpressionResultT")


class Operator(Token[ExpressionResultT]):
    """Subclass of token defining an operator."""

    __slots__: Tuple[str, ...] = tuple()

    @abc.abstractmethod
    def get_precedence(self) -> int:
        raise NotImplementedError


class InfixOperator(Operator[ExpressionResultT]):
    """Subclass of operator defining an infix operator."""

    __slots__: Tuple[str, ...] = tuple()

    @abc.abstractmethod
    def op(
        self, left: ExpressionResultT, right: ExpressionResultT
    ) -> ExpressionResultT:
        raise NotImplementedError


class PostfixOperator(Operator[ExpressionResultT]):
    """Subclass of operator defining an postfix operator."""

    __slots__: Tuple[str, ...] = tuple()

    @abc.abstractmethod
    def op(self, left: ExpressionResultT) -> ExpressionResultT:
        raise NotImplementedError


class Literal(Token[ExpressionResultT]):
    """Subclass of token defining a literal."""

    __slots__: Tuple[str, ...] = tuple()

    @abc.abstractmethod
    def val(self) -> ExpressionResultT:
        raise NotImplementedError


class RightParen(Token):
    """Subclass of token defining a right parenthesis."""

    __slots__: Tuple[str, ...] = tuple()

    def __repr__(self) -> str:
        return "<)>"


class LeftParen(Token):
    """Subclass of token defining a left parenthesis."""

    __slots__: Tuple[str, ...] = tuple()

    def __repr__(self) -> str:
        return "<(>"


def validate_tokens(token_list: List[Token]) -> None:
    """Validate the inputted tokens list (in infix ordering)."""

    token_list_prev: Iterable[Optional[Token]] = chain([None], token_list)

    paren_counter = 0

    for prev_token, curr_token in zip_longest(token_list_prev, token_list):
        # No postfix or infix operators at the beginning
        if prev_token is None and isinstance(
            curr_token, (InfixOperator, PostfixOperator)
        ):
            raise exceptions.InvalidRegexError(
                f"'{curr_token}' cannot appear at the start of a statement."
            )

        # No postfix operators at the end of a statement or right before another
        # operator or right paren
        elif isinstance(prev_token, InfixOperator):
            if curr_token is None:
                raise exceptions.InvalidRegexError(
                    f"'{prev_token}' cannot appear at the end of a statement."
                )
            elif isinstance(curr_token, (InfixOperator, PostfixOperator, RightParen)):
                raise exceptions.InvalidRegexError(
                    f"'{prev_token}' cannot appear immediately before '{curr_token}'."
                )

        # No left parens right before infix or postfix operators, or right
        # before a right paren
        elif isinstance(prev_token, LeftParen):
            if isinstance(curr_token, (InfixOperator, PostfixOperator)):
                raise exceptions.InvalidRegexError(
                    f"'{prev_token}' cannot appear immediately before '{curr_token}'."
                )

            # Track open/closed parens
            paren_counter += 1

        elif isinstance(prev_token, RightParen):
            paren_counter -= 1

            if paren_counter < 0:
                raise exceptions.InvalidRegexError(
                    "Token list has mismatched parethesis."
                )

    if paren_counter != 0:
        raise exceptions.InvalidRegexError("Token list has unclosed parethesis.")


def tokens_to_postfix(
    tokens: List[Token[ExpressionResultT]],
) -> List[Token[ExpressionResultT]]:
    """Takes in a list of tokens and changes them to postfix ordering."""

    stack: Deque[Token[ExpressionResultT]] = deque()
    res: List[Token[ExpressionResultT]] = []

    def comp_precedence(
        a: Token[ExpressionResultT], b: Token[ExpressionResultT]
    ) -> bool:
        """Compare precedence of operators (two tokens)."""
        return (
            cast(Operator[ExpressionResultT], a).get_precedence()
            <= cast(Operator[ExpressionResultT], b).get_precedence()
        )

    for c in tokens:
        if isinstance(c, Literal):
            res.append(c)
        elif isinstance(c, RightParen):
            while len(stack) > 0 and not isinstance(stack[-1], LeftParen):
                res.append(stack.pop())
            stack.pop()
        elif isinstance(c, LeftParen):
            stack.append(c)
        elif (
            not stack
            or isinstance(stack[-1], LeftParen)
            or not comp_precedence(c, stack[-1])
        ):
            stack.append(c)
        else:
            while (
                stack
                and not isinstance(stack[-1], LeftParen)
                and comp_precedence(c, stack[-1])
            ):
                res.append(stack.pop())
            stack.append(c)

    while stack:
        res.append(stack.pop())

    return res


def parse_postfix_tokens(
    postfix_tokens: List[Token[ExpressionResultT]],
) -> ExpressionResultT:
    """Parse list of postfix tokens to produce value of expression"""

    stack: Deque[ExpressionResultT] = deque()

    for token in postfix_tokens:
        if isinstance(token, InfixOperator):
            right = stack.pop()
            left = stack.pop()
            stack.append(token.op(left, right))
        elif isinstance(token, PostfixOperator):
            left = stack.pop()
            stack.append(token.op(left))
        elif isinstance(token, Literal):
            stack.append(token.val())
        else:
            raise exceptions.InvalidRegexError(f"Invalid token type {type(token)}")

    return stack[0]
